import pickle
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import auc, roc_curve, accuracy_score, recall_score
import matplotlib.pyplot as plt
from mksc.selection import model_method as mms
from mksc.selection import statistics_method as sms
from mksc.process import process as pp
from mksc.process import preprocess as pr
from mksc.process import woe_process as pw
import custom


def feature_engineering(feature, label):
    """
    特征工程过程函数：
    1. 特征组合
    2. 基于统计特性特征选择：缺失率、唯一率、众数比例
    3. 极端值处理
    4. TODO 异常值处理
    5. 缺失值处理
    6. COMMENT 归一化处理
    7. COMMENT 正态化处理
    8. 最优分箱
    9. IV筛选
    10. TODO PSI筛选
    11. 相关性筛选
    12. woe转化
    13. 逐步回归筛选

    Args:
        feature: 待处理特征数据框
        label: 特征数据框对应标签

    Returns:
        feature： 已完成特征工程的数据框
    """
    # 自定义特征组合模块
    feature = custom.Custom.feature_combination(feature)

    # 基于缺失率、唯一率、众数比例统计特征筛选
    missing_value = sms.get_missing_value(feature)
    distinct_value = sms.get_distinct_value(feature)
    unique_value = sms.get_unique_value(feature)
    feature.drop(set(missing_value['drop'] + distinct_value['drop'] + unique_value['drop']), axis=1, inplace=True)

    # 极端值处理
    feature, abnormal_value = pp.fix_abnormal_value(feature)

    # 缺失值处理
    feature, missing_filling = pp.fix_missing_value(feature)

    # 归一化处理
    # feature = pp.fix_scaling(feature)

    # 正态化处理
    # feature, standard_lambda = pp.fix_standard(feature)

    # 数值特征最优分箱，未处理的变量，暂时退出模型
    bin_result, iv_result, woe_result, woe_adjust_result = pw.binning(label, feature)
    bin_error_drop = bin_result['error'] + woe_adjust_result

    # IV筛选
    iv_drop = list(filter(lambda x: iv_result[x] < 0.02, iv_result))
    feature.drop(iv_drop + bin_error_drop, inplace=True, axis=1)

    # 相关性筛选
    cor_drop = sms.get_cor_drop(feature, iv_result)
    feature.drop(cor_drop, inplace=True, axis=1)

    # woe转化
    feature = pw.woe_transform(feature, woe_result, bin_result)

    # 逐步回归筛选
    feature_selected = mms.stepwise_selection(feature, label)
    feature = feature[feature_selected]

    # 中间结果保存
    result = {"missing_value": missing_value,
              "distinct_value": distinct_value,
              "unique_value": unique_value,
              "abnormal_value": abnormal_value,
              "missing_filling": missing_filling,
              "bin_result": bin_result,
              "iv_result": iv_result,
              "woe_result": woe_result,
              "woe_adjust_result": woe_adjust_result,
              "bin_error_drop": bin_error_drop,
              "iv_drop": iv_drop,
              "cor_drop": cor_drop,
              "feature_selected": feature_selected
              }
    with open('result/feature_engineering.pickle', 'wb') as f:
        f.write(pickle.dumps(result))
    return feature


def training(x_train, y_train, x_test, y_test, x_valid, y_valid):
    """
    模型训练过程函数
    1. 训练
    2. 预测
    3. 评估
    4. TODO 模型可选

    Args:
        x_train: 训练集特征
        y_train: 训练集标签
        x_test: 测试集特征
        y_test: 测试集标签
        x_valid: 验证集特征
        y_valid: 验证集标签
    """
    model = LogisticRegression()
    model.fit(x_train, y_train)

    # 预测结果
    predict_train = model.predict(x_train)
    predict_valid = model.predict(x_valid)
    predict_test = model.predict(x_test)

    # 模型评估
    acu_train = accuracy_score(y_train, predict_train)
    acu_valid = accuracy_score(y_valid, predict_valid)
    acu_test = accuracy_score(y_test, predict_test)

    sen_train = recall_score(y_train, predict_train, pos_label=1)
    sen_valid = recall_score(y_valid, predict_valid, pos_label=1)
    sen_test = recall_score(y_test, predict_test, pos_label=1)

    spe_train = recall_score(y_train, predict_train, pos_label=0)
    spe_valid = recall_score(y_valid, predict_valid, pos_label=0)
    spe_test = recall_score(y_test, predict_test, pos_label=0)
    print(f'模型准确率：验证 {acu_valid * 100:.2f}%	训练 {acu_train * 100:.2f}%	测试 {acu_test * 100:.2f}%')
    print(f'正例覆盖率：验证 {sen_valid * 100:.2f}%	训练 {sen_train * 100:.2f}%	测试 {sen_test * 100:.2f}%')
    print(f'负例覆盖率：验证 {spe_valid * 100:.2f}%	训练 {spe_train * 100:.2f}%	测试 {spe_test * 100:.2f}%')

    # K-s & roc
    predict_train_prob = np.array([i[1] for i in model.predict_proba(x_train)])
    fpr, tpr, thresholds = roc_curve(y_train.values, predict_train_prob, pos_label=1)
    auc_score = auc(fpr, tpr)
    w = tpr - fpr
    ks_score = w.max()
    ks_x = fpr[w.argmax()]
    ks_y = tpr[w.argmax()]
    fig, ax = plt.subplots()
    ax.plot(fpr, tpr, label='AUC=%.5f' % auc_score)
    ax.set_title('Receiver Operating Characteristic')
    ax.plot([0, 1], [0, 1], '--', color=(0.6, 0.6, 0.6))
    ax.plot([ks_x, ks_x], [ks_x, ks_y], '--', color='red')
    ax.text(ks_x, (ks_x + ks_y) / 2, '  KS=%.5f' % ks_score)
    ax.legend()
    fig.savefig("result/ks_roc.png")

    # 模型保存
    coefs = dict(list(zip(x_train.columns, list(model.coef_[0]))) + [("intercept_", model.intercept_[0])])
    with open('result/model.pickle', 'wb') as f:
        f.write(pickle.dumps(model))
    with open('result/coefs.pickle', 'wb') as f:
        f.write(pickle.dumps(coefs))


def main():
    """
    项目训练程序入口
    """
    # 加载数据、变量类型划分、特征集与标签列划分
    data = pr.load_data()
    numeric_var, category_var, datetime_var, label_var = pr.get_variable_type()
    feature = data[numeric_var + category_var + datetime_var]
    label = data[label_var]

    # 自定义数据清洗
    feature, label = custom.Custom.clean_data(feature, label)

    # 数据类型转换
    feature[numeric_var] = feature[numeric_var].astype('float')
    feature[category_var] = feature[category_var].astype('object')
    feature[datetime_var] = feature[datetime_var].astype('datetime64')

    # One-Hot编码
    feature = pd.concat([feature, pd.get_dummies(feature[category_var])], axis=1)
    feature.drop(category_var+datetime_var, inplace=True)

    # 特征工程
    feature = feature_engineering(feature, label)

    # 数据集划分
    x_train, x_valid_test, y_train, y_valid_test = train_test_split(feature, label, test_size=0.4, random_state=0)
    x_valid, x_test, y_valid, y_test = train_test_split(x_valid_test, y_valid_test, test_size=0.5, random_state=0)

    # 模型训练
    training(x_train, y_train, x_test, y_test, x_valid, y_valid)


if __name__ == "__main__":
    main()



import os

PATH = os.path.dirname(os.path.realpath(__file__))

with open(os.path.join(PATH, "configuration.ini"), "r", encoding='utf-8') as f:
    configuration_ini = f.read()

with open(os.path.join(PATH, "custom.py"), "r", encoding='utf-8') as f:
    custom_py = f.read()

with open(os.path.join(PATH, "train.py"), "r", encoding='utf-8') as f:
    train_py = f.read()

with open(os.path.join(PATH, "eda.py"), "r", encoding='utf-8') as f:
    eda_py = f.read()

with open(os.path.join(PATH, "score.py"), "r", encoding='utf-8') as f:
    score_py = f.read()

__all__ = [configuration_ini, custom_py, train_py, eda_py, score_py]



import sys
import os
from mksc.template import configuration_ini, custom_py, train_py, eda_py, score_py

def make_workspace(name):
    """
    创建项目的工作目录与脚本文件

    Args:
        name: 项目名
    """

    if os.path.exists(name):
        return f"Folder [{name}] is exists already, Please check out the work path"
    os.mkdir(name)
    os.mkdir(os.path.join(name, "conf"))
    os.mkdir(os.path.join(name, "data"))
    os.mkdir(os.path.join(name, "log"))
    os.mkdir(os.path.join(name, "model"))
    os.mkdir(os.path.join(name, "result"))

    with open(os.path.join(name, "custom.py"), "w", encoding='utf-8') as f:
        f.write(custom_py)
        
    with open(os.path.join(name, "eda.py"), "w", encoding='utf-8') as f:
        f.write(eda_py)

    configuration_ini_tmp = configuration_ini % (os.path.join(os.getcwd(), name),
                                                 os.path.join(os.getcwd(), name, "data"),
                                                 os.path.join(os.getcwd(), name, "result")
                                                 )
    with open(os.path.join(name, "conf", "configuration.ini"), "w", encoding='utf-8') as f:
        f.write(configuration_ini_tmp)

    with open(os.path.join(name, "train.py"), "w", encoding='utf-8') as f:
        f.write(train_py)        

    with open(os.path.join(name, "score.py"), "w", encoding='utf-8') as f:
        f.write(score_py)

def main():
    """
    命令行工具程序主入口
    """
    if len(sys.argv) == 1:
        return "CMD FORMAT: \n\tmksc project_name1 project_name2 ...\nPlease delivery one argument at least"
    else:
        for project_name in sys.argv[1:]:
            make_workspace(project_name)


if __name__ == "__main__":
    main()



import numpy as np
from scipy.stats import boxcox


def fix_abnormal_value(feature, threshold=0.05):
    """
    修正数据框中的数值型变量中的异常值

    Args:
        feature: 待修正的数据框
        threshold: 异常值替换阈值

    Returns:
        feature: 已处理数据框
        abnormal_value: 异常值统计结果
    """
    numeric_var = feature.select_dtypes(exclude=['object', 'datetime']).columns
    abnormal_value = {'result': {}, 'replace': []}
    for c in numeric_var:
        sm = feature[c].describe()
        iqr = sm['75%'] - sm['25%']
        min_ = sm['25%'] - 1.5*iqr
        max_ = sm['75%'] + 1.5*iqr
        abnormal_value_indexes = list(feature.loc[(feature[c] <= min_) | (feature[c] >= max_)].index)
        abnormal_value_length = len(abnormal_value_indexes)
        abnormal_value_rate = abnormal_value_length/feature[c].count()
        abnormal_value['result'][c] = {'abnormal_value_list': abnormal_value_indexes, 
                                       'abnormal_value_length': abnormal_value_length, 
                                       'abnormal_value_rate': abnormal_value_rate,
                                       'max': max_,
                                       'min': min_}
        if abnormal_value_rate <= threshold:
            abnormal_value['replace'].append(c)
            feature.loc[:, c] = feature.loc[:, c].apply(lambda x: x if (x < max_) & (x > min_) else np.nan)
    return feature, abnormal_value

def fix_missing_value(feature, threshold=0.05):
    """
    修正数据框中的数值型变量中的缺失值

    Args:
        feature: 待修正的数据框
        threshold: 缺失值替换阈值

    Returns:
        feature: 已处理数据框
        missing_filling: 缺失值统计结果
    """
    numeric_var = feature.select_dtypes(exclude=['object', 'datetime']).columns
    missing_filling = {'result': {}, 'replace': []}
    for c in feature:
        missing_rate = feature[c].isna().sum()/len(feature)
        if missing_rate <= threshold and missing_rate:
            missing_filling['result'][c] = {'fill_number': feature[c].mean(), 'missing_value_rate': missing_rate}
            missing_filling['replace'].append(c)
            if c in numeric_var:
                feature[c].fillna(missing_filling['result'][c], inplace=True)
            else:
                feature[c].fillna(feature[c].mode(), inplace=True)
    return feature, missing_filling

def fix_scaling(feature):
    """
    对数据框中的数据进行中性化处理

    Args:
        feature: 待处理的数据框

    Returns:
        feature: 已处理数据框
    """
    numeric_var = feature.select_dtypes(exclude=['object', 'datetime']).columns
    for c in numeric_var:
        sm = feature[c].describe()
        feature[c] = feature[c].apply(lambda x: (x - sm['mean'])/sm['std'] if x else x)
    return feature

def fix_standard(feature):
    """
    对数据框中的数据进行正态化处理

    Args:
        feature: 待处理的数据框

    Returns:
        feature: 已处理数据框
        standard_lambda: 对应特征的lambda值
    """
    numeric_var = feature.select_dtypes(exclude=['object', 'datetime']).columns
    standard_lambda = {}
    for c in numeric_var:
        feature[c], lambda_ = boxcox(feature[c]+0.5)
        standard_lambda[c] = lambda_
    return feature, standard_lambda
